package de.hft.stuttgart.citydoctor2.healing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.SolidNotClosedError;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.healing.math.EdgeGraph;
import de.hft.stuttgart.citydoctor2.math.Plane;
import de.hft.stuttgart.citydoctor2.math.Segment3d;
import de.hft.stuttgart.citydoctor2.math.Vector3d;

public class HealSolidNotClosedError implements HealingMethod {
	
	private static final Logger logger = LogManager.getLogger(HealSolidNotClosedError.class);

	private static double epsilon = 0.1;

	@Override
	public boolean visit(SolidNotClosedError err, ModificationListener l) {
		logger.debug("Executing Repair for SolidNotClosedError");
		if (fixVerticesCloseToEdges(err)) {
			// maybe errors are fixed don't try other fixes until next iteration
			return true;
		}
		return fixMissingPolygons(err);
	}

	private boolean fixMissingPolygons(SolidNotClosedError err) {
		if (err.getErrorEdges().size() <= 1) {
			// what to do with only one edge? Impossible?
			logger.error("{} only has one error edge, cannot fix", err.getGeometry());
			return false;
		}
		
		List<LinkedList<Edge>> candidatesForPolygon = new ArrayList<>();
		for (Edge e0 : err.getErrorEdges()) {
			for (Edge e1 : err.getErrorEdges()) {
				if (e1 == e0) {
					// don't check same edge
					continue;
				}
				Vertex v0 = e0.getConnectionPoint(e1);
				if (v0 != null) {
					// connected edges
					Vector3d ab = getVectorFromPoint(e0, v0);
					Vector3d ac = getVectorFromPoint(e1, v0);
					Vector3d norm = ab.cross(ac);
					Plane plane = new Plane(norm, v0);
					LinkedList<Edge> connectedEdges = new LinkedList<>();
					connectedEdges.add(e0);
					connectedEdges.add(e1);
					// collect all other edges which are in the same plane and connected
					collectEdgesConnectedAndInPlane(err, e0, e1, plane, connectedEdges);
					candidatesForPolygon.add(connectedEdges);
				}
			}
		}

		// build a polygon for the side where the most edges are available
		if (candidatesForPolygon.isEmpty()) {
			return false;
		}
		Collections.sort(candidatesForPolygon, (o1, o2) -> o2.size() - o1.size());
		LinkedList<Edge> candidate = candidatesForPolygon.get(0);
		// build a new polygon out of the connected edges
		buildPolygonFromEdges(err.getGeometry(), candidate);
		// have added a polygon, don't try to add more in this iteration
		return true;
	}

	private void buildPolygonFromEdges(Geometry geom, LinkedList<Edge> connectedEdges) {
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		LinearRing ext = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(ext);

		// edges form a loop?
		EdgeGraph graph = new EdgeGraph();
		for (Edge e : connectedEdges) {
			graph.addEdge(e, connectedEdges);
		}
		List<List<Edge>> cycles = graph.getCycles();

		// edges form a loop when nr edges >= nr vertices
		if (!cycles.isEmpty()) {
			// loop
			createRingFromLoop(cycles.get(0), ext);
		} else {
			// no cycle means open ring
			createRingFromNotClosedEdges(connectedEdges, ext);
		}
	}

	private void createRingFromNotClosedEdges(List<Edge> connectedEdges, LinearRing ext) {
		Edge firstEdge = connectedEdges.get(0);
		connectedEdges.remove(0);
		ext.addVertex(firstEdge.getFrom());
		ext.addVertex(firstEdge.getTo());
		Vertex currentVertex = firstEdge.getTo();
		boolean insertAtEnd = true;
		while (!connectedEdges.isEmpty()) {
			boolean addedEdge = false;
			for (Edge e : connectedEdges) {
				Vertex oppositeV = e.getOppositeVertex(currentVertex);
				if (oppositeV != null) {
					// found a connected edge
					// insert point depending where the edge is connected
					insertPointIntoRing(ext, insertAtEnd, oppositeV);
					currentVertex = oppositeV;
					connectedEdges.remove(e);
					addedEdge = true;
					break;
				}
			}
			if (!addedEdge) {
				if (!insertAtEnd) {
					// could not add more edges to lr, stop here
					// close ring
					ext.addVertex(ext.getVertices().get(0));
					return;
				}
				// no more edges in this direction, go the other direction
				insertAtEnd = false;
				currentVertex = firstEdge.getFrom();
			}
		}
		// close ring
		ext.addVertex(ext.getVertices().get(0));
	}

	private void insertPointIntoRing(LinearRing ext, boolean insertAtEnd, Vertex oppositeV) {
		if (insertAtEnd) {
			ext.addVertex(oppositeV);
		} else {
			// insert at the beginning
			ext.addVertex(0, oppositeV);
		}
	}

	private void createRingFromLoop(List<Edge> connectedEdges, LinearRing ext) {
		Edge firstEdge = connectedEdges.get(0);
		connectedEdges.remove(0);
		ext.addVertex(firstEdge.getFrom());
		ext.addVertex(firstEdge.getTo());
		Vertex currentVertex = firstEdge.getTo();
		while (!connectedEdges.isEmpty()) {
			boolean addedEdge = false;
			for (Edge e : connectedEdges) {
				Vertex oppositeV = e.getOppositeVertex(currentVertex);
				if (oppositeV != null) {
					ext.addVertex(oppositeV);
					currentVertex = oppositeV;
					connectedEdges.remove(e);
					addedEdge = true;
					break;
				}
			}
			if (!addedEdge) {
				return;
			}
		}
	}

	private void collectEdgesConnectedAndInPlane(SolidNotClosedError err, Edge e0, Edge e1, Plane plane,
			LinkedList<Edge> connectedEdges) {
		for (Edge e2 : err.getErrorEdges()) {
			if (e2 == e0 || e2 == e1) {
				// don't check the starting edges
				continue;
			}
			boolean connected = isConnectedToEdges(connectedEdges, e2);
			if (connected && plane.getDistance(e2.getTo()) < epsilon
					&& plane.getDistance(e2.getFrom()) < epsilon) {
				// edge is in plane and connected
				connectedEdges.add(e2);
			}
		}
	}

	private boolean isConnectedToEdges(List<Edge> connectedEdges, Edge e2) {
		for (Edge planeEdge : connectedEdges) {
			if (planeEdge.getConnectionPoint(e2) != null) {
				return true;
			}
		}
		return false;
	}

	private Vector3d getVectorFromPoint(Edge e0, Vertex v0) {
		Vector3d ab;
		if (v0 == e0.getTo()) {
			ab = e0.getFrom().minus(v0);
		} else {
			ab = e0.getTo().minus(v0);
		}
		return ab;
	}

	public static boolean fixVerticesCloseToEdges(SolidNotClosedError err) {
		boolean changed = false;
		Geometry geom = err.getGeometry();
		for (Edge e : err.getErrorEdges()) {
			Segment3d seg = new Segment3d(e.getFrom(), e.getTo());
			for (Vertex v : geom.getVertices()) {
				if (v == e.getFrom() || v == e.getTo()) {
					continue;
				}
				// check if there is an edge adjacent to the vertex that goes in the same
				// direction
				if (seg.getDistance(v) < epsilon && doParallelEdgesExist(geom, seg, v)) {
					addVertexToEdge(e, v);
					changed = true;
				}
			}
		}
		return changed;
	}

	private static boolean doParallelEdgesExist(Geometry geom, Segment3d seg, Vertex v) {
		Vector3d dir = seg.getPointB().minus(seg.getPointA());
		dir = dir.normalize();
		for (Edge e2 : geom.getEdgesAdjacentTo(v)) {
			Vector3d dir2 = e2.getTo().minus(e2.getFrom());
			dir2 = dir2.normalize();
			if (Math.abs(dir2.dot(dir)) > 0.99) {
				// at least one edge is parallel
				return true;
			}
		}
		return false;
	}

	private static void addVertexToEdge(Edge e, Vertex v) {
		for (LinearRing lr : e.getAdjacentRings()) {
			for (int i = 0; i < lr.getVertices().size() - 1; i++) {
				Vertex v1 = lr.getVertices().get(i);
				Vertex v2 = lr.getVertices().get(i + 1);
				if (e.formedByIgnoreDirection(v1, v2)) {
					lr.addVertex(i + 1, v);
					break;
				}
			}
		}
	}

	@Override
	public HealSolidNotClosedError createNew() {
		return new HealSolidNotClosedError();
	}

	@Override
	public HealingID getID() {
		return HealingID.S_NOT_CLOSED;
	}
	
}
